﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Boardmodul: Bereitstellung gerätespezifischer Funktionen und Mechanismen            //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2018 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "AVRChip.h"
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Definitionen  ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define SYS_LED     A,7
#define BUS_LED     C,6

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef enum                        // Arbeitsanweisungen für das Board-Modul:
{   DO_STARTUP,                     // · Starte die Anwendung neu
    DO_ENTER_PGMLDR,                // · Starte den Programmlader
    DO_PROCEED,                     // · Führe die Arbeit fort
    DO_SHUTDOWN                     // · Schalte das Board ab
} DoNext;

typedef struct                      // Bus-Übertragungsstatus:
{   bool   pending : 1;             // · Übertragung im Gange
    bool   echo    : 1;             // · Echo erhalten
    bool   error   : 1;             // · Übertragung Fehlgeschlagen (Kollision)
    UInt08 byteGap : 5;             // · Vergangene Zeit seit dem letzten Byteempfang [ms]
} BusStatus;

typedef struct                      // Element der Bus-Empfangsschlange:
{   bool   timeout : 1;             // · Zeitüberschreitung bei 'byteGap' aufgetreten
    UInt08 padding : 7;             // · Füllbits (ohne Bedeutung)
    UInt08 data    : 8;             // · Datenbyte
} BusRxItem;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static volatile UInt16       msecTicks                             = 0;
static volatile UInt16       secTicks                              = 0;
static volatile UInt08       busTxQueue[CFG_BUS_TX_QUEUE_CAPACITY] = { 0 };
static volatile UInt08       busTxQueueHead                        = 0;
static volatile UInt08       busTxQueueLength                      = 0;
static volatile UInt08       busLoopbackByte                       = 0;
static volatile BusStatus    busStatus                             = { false, false, false, 0};
static volatile BusRxItem    busRxQueue[CFG_BUS_RX_QUEUE_CAPACITY] = { { 0, false, false } };
static volatile UInt08       busRxQueueHead                        = 0;
static volatile UInt08       busRxQueueTail                        = 0;
static          DoNext       doNext           				       = DO_STARTUP;
static          UInt08       busActLedOn                           = 0;
static          bool         heartbeatVisible                      = true;

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  Reset(): Führt einen physikalischen Reset herbei indem der Watchdog ausgelöst wird. //
//  -------  Reset() kehrt nie zum Aufrufer zurück.                                     //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((noreturn)) static void
Reset(void)
{
    DisableInterrupts();            // Interrupts sperren
    wdt_enable(WDTO_15MS);          // Watchdog-Intervall auf kürzeste Zeit setzen
    __asm__ volatile
    (   "ldi r30, lo8 (%0) \n"      // Nach einem Neustart des Systems wird als
	    "ldi r31, hi8 (%0) \n"      // erstes das Programmladermodul ausgeführt.
	    "ldi r24, %1       \n"      // Der Programmlader nutzt zur Steuerung seines
	    "ldi r25, hi8 (%2) \n"      // Ablaufs Variablen, die in der "Noinit"-Sektion
	    "0:                \n"      // angesiedelt sind, d.h. sie werden im Rahmen
	    "st  Z+,  r24      \n"      // eines Neustartes nicht initialisiert.
	    "cpi r30, lo8 (%2) \n"      // Um nun zu vermeiden, dass übrig gebliebene
        "cpc r31, r25      \n"      // Werte des SRAMs unerwünschte Effekte im
	    "brlo 0b"                   // Programmlader hervorrufen, wird vor dem
            :                       // Reset das komplette SRAM mit Nullen überschrieben.
	        : "i" (RAMSTART),       // Da durch die Nullung Heap wie Stack komplett
              "i" (0),              // zerstört sind, ist diese Aktion ein Punkt-Ohne-
              "i" (RAMEND+1)        // Wiederkehr.
	);
    while (true);                   // Warten, bis Watchdog-Reset eintritt
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
EnterPgmLdr(void)
{
    if (pgm_read_word(FLASHEND - CFG_PGMLDR_SIZE*1024 + 1) == 0x940C)
    {   GOTO(FLASHEND - CFG_PGMLDR_SIZE*1024+1 + 4);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_PreStart(): Diese Funktion wird einmal nach dem Systemstart aufgerufen. Der   //
//  ----------------  Zeitpunkt ihres Aufrufs macht sie zur einer Sonderfunktion, denn  //
//  ihr Aufruf findet in der Start-Sektion "init3", noch lange vor Aufruf von main(),   //
//  statt. Das ist der richtige Zeitpunkt, um z.B. die Ursache des Resets zu ermitteln  //
//  oder statische Variablen auszulesen, deren Werte vor dem Reset platziert wurden     //
//  und den Neustart überlebt haben. Denn zu diesem Zeitpunkt sind die statischen       //
//  Variablen noch nicht initialisiert.                                                 //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
__attribute__((naked, section(".init3"))) void
BOARD_PreStart(void)
{
    MCUSR = 0;
    wdt_disable();
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Interruptbehandler  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
ISR(TIMER1_COMPA_vect)
{
    // Ticker inkrementieren
    {   msecTicks++;
        static UInt16 counter = 0;
        counter++;
        if (counter == 1000)
        {   counter = 0;
            secTicks++;
        }
    }
    // Bus Byte-Zeitlücke
    if (busStatus.byteGap <= CFG_BUS_INTERBYTE_TIMEOUT)
    {   busStatus.byteGap++;
        if (busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT)
        {   busRxQueue[busRxQueueTail].timeout = true;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART_TXC_vect)
{                                   // IRQ Abfolge sowie die Signale 'echo' und 'pending':
    if (!busStatus.echo)            //
    {   busStatus.error = true;     // Start Tx:    |1|
        busTxQueueLength = 0;       //               :
    }                               // Tx-IRQ:  ···--:-----|2|----|3|-···-|n|----| |-···
    busStatus.echo = false;         // Rx-IRQ:  ···--:-|1|--:-|2|--:--···-- -|n|--:--···
    UInt08 head = busTxQueueHead;   //               :  :___:  :___:          :___:
    if (busTxQueueLength == 0)      // echo:    ...__:__|   |__|   |____...___|   |__...
    {   busStatus.pending = false;  //               :____________________________:
    }                               // pending: ...__|                            |__...
    else
    {   busLoopbackByte = busTxQueue[head];
        UDR             = busTxQueue[head];
        head++;
        if (head == CARDINALITY(busTxQueue)) head = 0;
        busTxQueueLength--;
    }
    busTxQueueHead = head;
}

//////////////////////////////////////////////////////////////////////////////////////////
ISR(USART_RXC_vect)
{
    UInt08 byte = UDR;
    busStatus.byteGap = 0;   // Byte empfangen, Zeitzähler zurücksetzen
    if (busStatus.pending && byte == busLoopbackByte) busStatus.echo = true; // Echo
    // Byte in Empfangsschlange stellen
    UInt08 tail = busRxQueueTail;
    busRxQueue[tail].data = byte;
    tail++;
    if (tail == CARDINALITY(busRxQueue)) tail = 0;
    if (tail == busRxQueueHead ) busRxQueue[busRxQueueTail].timeout = true; // Überlauf
    else                         busRxQueueTail = tail;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Funktionen ////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_Reset(): Initiiert einen Systemneustart. Die Funktion löst nicht unmittelbar  //
//  -------------  einen Reset aus, sondern führt zu einem Ausstieg aus der Haupt-      //
//  schleife von Main(). Die danach sukzessiv aufgerufenen "Shutdown"-Funktionen aller  //
//  betroffenen Module bewirken den kontrollierten Abschluss aller Aktivitäten, bevor   //
//  die letzte aufgerufene "Shutdown"-Funktion des Board-Moduls dann den physikali-     //
//  schen Reset auslöst.                                                                //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Reset(void)
{
    doNext = DO_STARTUP;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EnterPgmLdr(): Ähnlich wie BOARD_Reset() initiiert BOARD_EnterPgmLdr() einen  //
//  -------------------  Systemneustart. Im Unterschied zu BOARD_Reset(), wird hier     //
//  jedoch nicht die Anwendung neu gestartet, sondern das System bootet in das Pro-     //
//  grammladermodul.                                                                    //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EnterPgmLdr(void)
{
    doNext = DO_ENTER_PGMLDR;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_MSecTicks(): Die Funktion liefert den Stand eines monotonen Millisekunden-    //
//  -----------------  Zeitzählers. Der Zähler wird interruptgetrieben inkrementiert    //
//  und hat eine Breite von 16-Bit. Kommt es nach 65535 ms zum Überlauf, beginnt der    //
//  Zeitzähler wieder bei 0.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Millisekunden-Zeitzählers.                             //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_MSecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = msecTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_SecTicks(): Die Funktion liefert den Stand eines monotonen Sekunden-Zeit-     //
//  ----------------  zählers. Der Zähler wird interruptgetrieben inkrementiert und hat //
//  eine Breite von 16-Bit. Kommt es nach 65535 s (knapp 18¼h) zum Überlauf, beginnt    //
//  der Zeitzähler wieder bei 0.                                                        //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: Aktueller Stand des Sekunden-Zeitzählers.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
BOARD_SecTicks(void)
{
    UInt16 ticks = 0;

    Atomic { ticks = secTicks; }
    return ticks;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPLoad(): Kopiert 'count' Bytes aus dem EEPROM-Speicher ins RAM.             //
//  ---------------                                                                     //
//                                                                                      //
//  Eingabe: · eepAddress:  Adresse im EEPROM-Speicher.                                 //
//           · ramAddress:  Adresse im Ram.                                             //
//           · count:       Anzahl zu kopierender Bytes.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPLoad(const void* eepAddress, void* ramAddress, UInt16 count)
{
    eeprom_read_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_EEPSave(): Kopiert 'count' Bytes aus dem RAM in den EEPROM-Speicher. Da       //
//  ---------------  EEPROM-Speicher lt. Datenblatt nur ca. 100000 Lösch/Schreibzyklen  //
//  pro Zelle verträgt, wird vor für jedes Bytes vor dem Speichervorgang geprüft, ob    //
//  sich der zu schreibende Wert sich vom Inhalt der EEP-Speicherzelle unterscheidet.   //
//  Erst danach wird nötigenfalls ein Lösch- und Schreibvorgang durchgeführt.           //
//  Hinweis: Man beachte, dass das Beschreiben einer EEPROM-Speicherzelle ca. 8 ms in   //
//           Anspruch nimmt und damit verhältnismäßig viel Zeit verschlingt.            //
//                                                                                      //
//  Eingabe: · ramAddress:  Adresse im Ram.                                             //
//           · eepAddress:  Adresse im EEPROM-Speicher.                                 //
//           · count:       Anzahl zu kopierender Bytes.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_EEPSave(const void* ramAddress, void* eepAddress, UInt16 count)
{
    eeprom_update_block(ramAddress, (void*)eepAddress, count);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsIdle(): Prüft, ob der Bus momentan frei ist. Der Bus gilt als frei, wenn //
//  -----------------  eine Kommunikationspause von mehr als CFG_BUS_INTERBYTE_TIMEOUT  //
//  Millisekunden eintritt.                                                             //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Bus frei ist.                                             //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsIdle(void)
{
    bool  idle = false;
    Atomic
    {   idle = !busStatus.pending && busStatus.byteGap > CFG_BUS_INTERBYTE_TIMEOUT;
    }
    return idle;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTransmit(): Sendet das Datenbyte in 'byte' auf den Bus.                    //
//  -------------------  Die Sendeeinheit ist durch folgenden Zustandsautomaten         //
//  realisiert, Ausgangszustand ist IDLE:                                               //
//                                                                                      //
//                                                                                      //
//     +----------+ BOARD_BusTransmit()    +------------------+ BOARD_BusTransmit()     //
//     |//////////+----------------------->|//////////////////+--------------------+    //
//     |// IDLE //|              COMPLETED |// TRANSMITTING //|                    |    //
//     |//////////|<-----------------------+//////////////////|<-------------------+    //
//     +----------+                        +--------+---------+                         //
//          /|\                                     | COLLISION                         //
//           |                                      |                                   //
//           |                                     \|/                                  //
//           |                               +--------------+                           //
//           | BOARD_BusTakeFailure()        |//////////////|                           //
//           +-------------------------------+// FAIILURE //|                           //
//                                           |//////////////|                           //
//                                           +--------------+                           //
//                                                                                      //
//  Mit Aufruf dieser Funktion wechselt der Automat IDLE nach TRANSMITTING und beginnt  //
//  mit der Übertragung des angegebenen Datenbytes. Ist bereits eine Übertragung im     //
//  Gange (Zustand TRANSMITTING), wird bei Aufruf der Funktion das zu übertragende      //
//  Datenbyte für die spätere Übertragung in die Sende-Warteschlange gestellt. Nach     //
//  Aussendung aller zu übertragenden Datenbytes (Ereignis COMPLETED) kehrt der Automat //
//  wieder in den Zustand IDLE zurück.                                                  //
//  Das Halbduplexverfahren des seriellen Busses bedingt, dass jedes ausgesendete Byte  //
//  unmittelbar als Echo wieder in den Empfangspuffer zurückwandert. Diese Rücküber-    //
//  tragung wird genutzt, um Kollisionen aufzudecken. Unterscheiden sind ausgesendetes  //
//  Byte und Echo, löst das das Ereignis COLLISION aus und damit den Wechsel in den     //
//  Zustand FAIILURE. Die Übertragung wird in dem Fall nicht mehr fortgesetzt und die   //
//  Sende-Warteschlange geleert. Gegebenenfalls noch ausstehende Bytes werden also      //
//  nicht mehr übertragen. Im Zustand FAIILURE bleiben Aufrufe von BOARD_BusTransmit()  //
//  ohne Wirkung. Um neue Datenbytes senden zu können, muss der Automat erst durch      //
//  Aufruf von BOARD_BusTakeFailure() wieder in den Zustand IDLE versetzt werden.       //
//                                                                                      //
//  Eingabe: · byte: Das zu übertragende Datenbyte.                                     //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusTransmit(UInt08 byte)
{
    // Ist die Übertragungsschlange voll, warten, bis wieder Platz ist
    bool  queueFull = false;
    do
    {   Atomic
        {   queueFull = busTxQueueLength == CARDINALITY(busTxQueue);
	    }
    } while (queueFull);
    // Byte senden
    Atomic
    {   if (busStatus.error) return; // Zustand FAIILURE: Nichts tun
        if (busStatus.pending)
        {   // Zustand TRANSMITTING: Byte für Übertragung aufreihen
            UInt08 tail = busTxQueueHead + busTxQueueLength;
            if (tail >= CARDINALITY(busTxQueue)) tail -= CARDINALITY(busTxQueue);
            busTxQueue[tail] = byte;
            busTxQueueLength++;
        }
        else
        {   // Zustand IDLE: Byte unmittelbar senden
            busStatus.pending = true;
            busStatus.echo    = false;
            busLoopbackByte   = byte;
            UDR               = byte;
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsTransmitting(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit())   //
//  -------------------------  im Zustand TRANSMITTING befindet. Das ist solange der    //
//  Fall, wie die Sendeeinheit mit der Übertragung von Datenbytes beschäftigt ist.      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn der Sendevorgang noch im Gange ist.                           //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsTransmitting(void)
{
    bool  pending = false;
    Atomic
    {   pending = busStatus.pending;
    }
    return pending;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusTakeFailure(): Prüft, ob sich der Automat (siehe BOARD_BusTransmit()) im   //
//  ----------------------  Zustand FAILURE befindet und löst gegebenenfalls einen      //
//  Wechsel in den Zustand IDLE aus. Übertragungsfehler entstehen in der Regel durch    //
//  Kollisionen auf dem Bus.                                                            //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', wenn ein Übertragungsfehler vorlag.                                //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusTakeFailure(void)
{
    bool  error = false;
    Atomic
    {   error = busStatus.error;
        busStatus.error = false;
    }
    return error;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusReceive(): Steht bei Aufruf der Funktion ein empfangenes Datenbyte zur     //
//  ------------------  Abholung bereit, wird es aus der Empfangsschlange entfernt und  //
//  in 'byte' übergeben. BOARD_BusReceive() kehrt so lange mit 'true' zurück, wie       //
//  Datenbytes zur Abholung bereitstehen.                                               //
//  Man beachte: Der Rückgabewert von BOARD_BusReceive() sagt nichts darüber aus, ob    //
//  eine Empfangssequenz noch im Gange ist oder nicht. Dazu siehe BOARD_BusIs-          //
//  Receiving().                                                                        //
//                                                                                      //
//  Eingabe: · byte: Adresse des Bytes, das das nächste zur Abholung bereitstehende     //
//                   Datenbyte aufnimmt.                                                //
//  Ausgabe: 'true', wenn in 'byte' ein Datenbyte abgelegt wurde.                       //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusReceive(UInt08* byte)
{
    bool  hasData = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        if (!busRxQueue[head].timeout)
        {   // Keine Zeitüberschreitungs-Markierung gesetzt
            if (head != busRxQueueTail)
            {   // Datenbyte vorhanden
                *byte = busRxQueue[head].data;
				head++;
				if (head == CARDINALITY(busRxQueue)) head = 0;
				busRxQueueHead = head;
                hasData = true;
			}
		}
    }
    return hasData;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusIsReceiving(): Prüft, ob die aktuelle Empfangssequenz noch im Gange ist.   //
//  ----------------------                                                              //
//  Eine Gruppe empfangener Datenbytes gilt als zusammenhängene Empfangssequenz, wenn   //
//  die Zeitspanne zwischen dem Empfang sukzessiver Datenbytes einen bestimmten Zeit-   //
//  rahmen nicht überschreitet. Mit anderen Worten: Liegen die Zeitpunkte zweier auf-   //
//  einanderfolgend empfangener Datenbytes länger als CFG_BUS_INTERBYTE_TIMEOUT Milli-  //
//  sekunden auseinander (timeout), gilt die Empfangssequenz als unterbrochen.          //
//                                                                                      //
//    Empfang: ···-|A|-|B|-|C|---------------×---|D|----|E|---------------×--|F|-···    //
//                           :               :            :               :             //
//                           |--- timeout ---|            |--- timeout ---|             //
//                                                                                      //
//	  Empfangssequenzen: ABC, DE, F... 	                                                //
//                                                                                      //
//  Die Unterbrechungen werden mit in der Empfangsschlange vermerkt, so dass beim Aus-  //
//  lesen der empfangenen Bytes mittels BOARD_BusReceive() die Empfangssequenzen zeit-  //
//  unabhängig separiert werden. Solange BOARD_BusIsReceiving() mit 'true' zurückkehrt, //
//  ist die Empfangssequenz noch im Gange.                                   			//
//                                                                                      //
//      Beispiel: Empfang von Byte-Sequenzen:                                           //
//                                                                                      //
//       UInt08 sequence[SEQUENCE_CAPACITY];   // SEQUENCE_CAPACITY: 1 .. 255           //
//                                                                                      //
//       UInt08                                                                         //
//       TakeNextSequence(void)                                                         //
//       {                                                                              //
//           static UInt08 index = 0;                                                   //
//                                                                                      //
//           UInt08 length = 0;                                                         //
//           if (BOARD_BusIsReceiving())                                                //
//           {   // Empfang einer Sequenz im Gange                                      //
//               UInt08 byte;                                                           //
//               while (BOARD_BusReceive(&byte)) // Datenbytes abholen und aufsammeln   //
//               {   if (index < SEQUENCE_CAPACITY)                                     //
//                   {   sequence[index] = byte;                                        //
//                       index++;                                                       //
//                   }                                                                  //
//                   else                                                               //
//                   {   // Überlauf ...                                                //
//                   }                                                                  //
//               }                                                                      //
//           }                                                                          //
//           else                                                                       //
//           {    // Empfang einer Sequenz nicht (mehr) im Gange                        //
//                length = index;                                                       //
//                index  = 0;                                                           //
//           }                                                                          //
//           return length;   // Rückgabewert: Aktuelle Sequenzlänge                    //
//       }                                                                              //
//                                                                                      //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: 'true', solange die Empfangssequenz im Gange ist.                          //
//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_BusIsReceiving(void)
{
    bool  timeout = false;
    Atomic
    {   UInt08 head = busRxQueueHead;
        timeout = busRxQueue[head].timeout;
        if (head != busRxQueueTail) busRxQueue[head].timeout = false;
    }
    return !timeout;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  BOARD_BusOnActivity(): Rückruffunktion der Bus-Übertragungseinheit. Die Funktion    //
//  ---------------------  wird aufgerufen, wenn mit dem Gerät über den Bus             //
//  kommuniziert wird.                                                                  //
//                                                                                      //
//  Eingabe: -                                                                          //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_BusOnActivity(void)
{
    busActLedOn = CFG_BUS_ACT_LED_FLASH_DURATION;    
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_SetHeartbeatVisible(bool visible)
{
    heartbeatVisible = visible;
    if (!visible) ClrPin(SYS_LED);
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_HeartbeatVisible(void)
{
    return heartbeatVisible;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Startup(void)
{
    if (doNext != DO_STARTUP) Reset();
    DisableInterrupts();
    // Watchdog --------------------------------------------------------------------------
    wdt_enable(WDTO_2S);
    // Ports'n'Pins ----------------------------------------------------------------------
    ConfPort(A, _T_,    // PA.0: A0_GPIO    : Allgemeiner Anschluss
                _T_,    // PA.1: A1_GPIO    : Allgemeiner Anschluss
                _T_,    // PA.2: A2_GPIO    : Allgemeiner Anschluss
                _T_,    // PA.3: A3_GPIO    : Allgemeiner Anschluss
                _T_,    // PA.4: A4_PDN	    : Allgemeiner Pull-Down-Anschluss
                _T_,    // PA.5: A5_PUP     : Allgemeiner Pull-Up-Anschluss
                _T_,    // PA.6: A6_AIN     : Analogeingang 0 .. 5V
                _L_);   // PA.7: A7_SYS_LED : System-Led: L = aus
    ConfPort(B, _T_,    // PB.0: B0_GPIO    : Allgemeiner Anschluss
                _T_,    // PB.1: B1_TA1     : Allgemeine Drucktaste 1, L = gedrückt
                _T_,    // PB.2: B2_TA2     : Allgemeine Drucktaste 2  L = gedrückt
                _T_,    // PB.3: B3_PIEZO   : Piezosignalgeber, Frequenz über Timer 0 PWM
                _T_,    // PB.4: B4_GPIO    : Allgemeiner Anschluss
                _T_,    // PB.5: B5_MOSI    : SPI/ISP
                _T_,    // PB.6: B6_MISO    : SPI/ISP
                _T_);   // PB.7: B7_SCK     : SPI/ISP
    ConfPort(C, _T_,    // PC.0: C0_SCL     : TWI (I²C)
                _T_,    // PC.1: C1_SDA     : TWI (I²C)
                _T_,    // PC.2: C2_TCK     : JTAG
                _T_,    // PC.3: C3_TMS     : JTAG
                _T_,    // PC.4: C4_TDO     : JTAG
                _T_,    // PC.5: C5_TDI     : JTAG
                _L_,    // PC.6: C6_BUS_ACT : Cy4NET-Aktivität, L = keine
                _T_);   // PC.7: C7_GPIO    : Allgemeiner Anschluss
    ConfPort(D, _T_,    // PD.0: RXD        : Bus-RXD
                _T_,    // PD.1: TXD        : Bus-TXD
                _T_,    // PD.2: D2_GPIO    : Allgemeiner Anschluss
                _T_,    // PD.3: D3_GPIO    : Allgemeiner Anschluss
                _T_,    // PD.4: D4_GPIO    : Allgemeiner Anschluss
                _T_,    // PD.5: D5_GP_LED3 : Allgemeine LED 3, H = Ein
                _T_,    // PD.6: D6_GP_LED2 : Allgemeine LED 2, H = Ein
                _T_);   // PD.7: D7_GP_LED1 : Allgemeine LED 1, H = Ein (PWM Timer 2)
    // Pause, bis sich alles stabilisiert hat --------------------------------------------
    Freeze100us();
    // Timer 1: Clear Timer on Compare (CTC), Intervall: 1ms (Wiederholrate: 1kHz) -------
    TCCR1B = BV(WGM12) | BV(CS10);
    OCR1A  = F_CPU/1000-1;
    TIMSK |= BV(OCIE1A);
    // USART 19200 Baud, 8 Data, 1 Stop, No Parity, Async., IRQs: RXC/TXC ----------------
    UCSRB = BV(RXCIE) | BV(TXCIE) | BV(RXEN) | BV(TXEN);
    UBRRH = (F_CPU/16L/19200-1)>>8;
    UBRRL = (F_CPU/16L/19200-1);
    // Sonstiges -------------------------------------------------------------------------
    EnableInterrupts();
    doNext = DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
bool
BOARD_DoMainLoop(void)
{
    return doNext == DO_PROCEED;
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Proceed(void)
{
    // 1ms Zeitbarriere
	UInt16 ticks = BOARD_MSecTicks();
	{	static UInt08 ms = 0;
		if (ms == (UInt08)ticks) return;
		ms = (UInt08)ticks;
	}
    wdt_reset();
    // Herzschlag-LED
    if (heartbeatVisible)
    {   ChgPin(SYS_LED, (bool)((ticks&0x0380)==0x0080 || (ticks&0x0380)==0x0180));
    }
    // Bus-Aktivitäts-LED
    {   if (busActLedOn > 0)
        {   busActLedOn--;
            ChgPin(BUS_LED, busActLedOn > 0);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
BOARD_Shutdown(void)
{
	while (BOARD_BusIsTransmitting());
    switch( doNext )
    {   case DO_PROCEED:      doNext = DO_SHUTDOWN; return;
        case DO_ENTER_PGMLDR: EnterPgmLdr();        break;
        case DO_STARTUP:      Reset();              break;
        default:              ;
    }
    // Abschaltanweisungen ...
}

//////////////////////////////////////////////////////////////////////////////////////////
